-- Finds the combined length of all [[sub]sub]field
-- values.  This is just an example; this may not be
-- how you want to define fields's lengths.
local findLength
function findLength(Node)
   if Node:isLeaf() then
      return #Node:nodeValue()
   else
      local TotalLength = 0
      for i=1,#Node do
         TotalLength = TotalLength + findLength(Node[i])
      end
      return TotalLength
   end
end

local ExpectedDateTimeFormats = {
   {'^(%d%d%d%d)(%d%d)(%d%d)$', -- pattern with captures
      {{1899,2050},{1,12},{1,31}}}, -- valid range of number for each capture
   
   {'^(%d%d%d%d)(%d%d)(%d%d)(%d%d)(%d%d)(%d%d)[%.%-%d]*$', -- pattern with captures
      {{1899,2050},{1,12},{1,31},{0,24},{0,59},{0,59}}} -- valid range of number for each capture
}

local function isDate(Field)
   local S = tostring(Field)
   for i,F in ipairs(ExpectedDateTimeFormats) do
      local Captures = {S:match(F[1])}
      if #Captures > 1 then
         local InRange = true
         for j,C in ipairs(Captures) do
            local N = tonumber(C)
            local R = F[2][j]
            InRange = InRange and
                      R[1] <= N and N <= R[2]
         end
         if InRange then return true end
      end
   end
   return false
end

-- For use with the "Field Type" custom property.
-- Ordered from most specific to least specific.
local FieldTypeFunctions = {
   {'Date', isDate},
   {'Phone Number', function(Field)
         return nil ~= tostring(Field):match('^%(?%d%d%d%)?%-?%d%d%d%-?%d%d%d%d$')
      end},
   {'Numeric', function(Field)
         local S = tostring(Field)
         return S:match('^%d*%.?%d+$') or
                S:match('^%d+%.?%d*$')
      end},
   {'Composite', function(Field, Segment, Index)
         if (Segment:nodeName() == 'MSH') and (Index < 3) then
            return false
         else
            return tostring(Field):match('[%^&]')
         end
      end},
   {'Code', function(Field)
         return tostring(Field):match('^%u%u?%u?$') -- 1-3 uppercase letters
      end},
   {'Text', function(Field)
         return true -- anything matches
      end}
}

local function guessSingleFieldType(Field, Segment, Index)
   for i,FT in ipairs(FieldTypeFunctions) do
      if FT[2](Field, Segment, Index) then
         return tostring(i)
      end
   end
   return #FieldTypeFunctions
end

local function guessOverallFieldType(ResultList)
   for i,V in ipairs(ResultList) do
      ResultList[i] = tonumber(ResultList[i])
   end
   local FTi = #FieldTypeFunctions
   if #ResultList > 0 then
      FTi = math.min(unpack(ResultList))
   end
   return FieldTypeFunctions[FTi][1]
end

local function bestSampleValue(ResultList)
   -- TODO scrub data?  How to know which ones to scrub?
   local Best = ''
   local BestSubs = 0
   local BestLength = 0
   for i,S in ipairs(ResultList) do
      local Subs,Orig = S:match('(%d+),(.*)')
      Subs = tonumber(Subs)
      if Subs > BestSubs or (Subs == BestSubs and #Orig > BestLength) then
         Best = Orig
         BestSubs = Subs
         BestLength = #Orig
      end
   end
   return Best
end

-- Add/modify custom properties.
ifdoc.customProperties = {
   {
      name = 'Field Type',
      resultType = 'string',
      fieldFunc = guessSingleFieldType,
      finalResultFunc = guessOverallFieldType
   },
   {
      name = 'Subfields',
      resultType = 'number',
      fieldFunc = function(Field)
         if Field:isLeaf() then
            return nil -- N/A
         else
            local MaxSubField = 1
            for i=2,#Field do
               if not Field[i]:isNull() then
                  MaxSubField = i
               end
            end
            return (MaxSubField > 1 and MaxSubField) or nil
         end
      end,
      finalResultFunc = function(ResultList)
         if #ResultList > 0 then
            return math.max(unpack(ResultList))
         else
            return nil -- N/A
         end
      end
   },
   {
      name = 'Sample value', -- will actually return the longest value
      resultType = 'string',
      fieldFunc = function(Field)
         return #Field..','..tostring(Field)
      end,
      finalResultFunc = bestSampleValue
   },
   --[[
   {
      name = 'Maximum length',
      resultType = 'number',
      fieldFunc = findLength,
      finalResultFunc = function(ResultList)
         if #ResultList > 0 then
            return math.max(unpack(ResultList))
         else
            return nil -- N/A
         end
      end
   },
   --]]
}
