ifdoc.page = {}

local function getRequiredParameter(Params, N)
   if Params[N] then
      return Params[N]
   else
      error('Missing required parameter \''..N..'\'', 2)
   end
end

local function getProgress(Params)
   local ProjectGuid = getRequiredParameter(Params, 'project')
   if not ProjectGuid then
      error('Missing required parameter: project')
   end
   ifdoc.init(ProjectGuid)
   
   local R = {complete=true}
   
   local CRS = ifdoc.db.execute('SELECT DataSize FROM InProgress')
   if #CRS ~= 0 and not CRS[1].DataSize:isNull() then
      R.complete = false
      R.total = tonumber(CRS[1].DataSize:nodeValue())
      local MRS = ifdoc.db.execute('SELECT COUNT(DISTINCT Id) AS C FROM Message')
      R.processed = (not MRS[1].C:isNull() and tonumber(MRS[1].C:nodeValue())) or 0
   end
   
   local RJson = json.serialize{data=R}
   net.http.respond{entity_type='text/json', body=RJson}
end

local function generateData(Params)
   local ProjectGuid = getRequiredParameter(Params, 'project')
   if not ProjectGuid then
      error('Missing required parameter: project')
   end
   ifdoc.PROJECT_GUID = ProjectGuid
   ifdoc.destroy()
   ifdoc.init(ProjectGuid)
   
   local SD = ifdoc.getSampleData(ProjectGuid)
   
   ifdoc.db.execute('DELETE FROM InProgress')
   ifdoc.db.execute('INSERT INTO InProgress VALUES('..#SD..')')
   
   -- This will send the response before we enqueue the messages
   getProgress(Params)
   
   for i=1,#SD do
      local D = SD[i].Message:nodeValue()
      local Q = {['ProjectGuid']=ProjectGuid, ['Data']=D, ['SampleDataIndex']=i}
      local QJson = json.serialize{data=Q}
      queue.push{data=QJson}
   end
   
   local Done = {['ProjectGuid']=ProjectGuid, ['Complete']=true}
   local DoneJson = json.serialize{data=Done}
   queue.push{data=DoneJson}
end

TrackedTimes = {}
function trackTime(Name, Start, Finish)
   TrackedTimes[Name] = (TrackedTimes[Name] or 0) + os.difftime(Start, Finish or os.time())
end

local function renderTemplateFile(FileName, Env)
   local TemplatePath = iguana.project.root()..'other/ifdoc/'..FileName
   local F = io.open(TemplatePath, 'r')
   local T = ((F and F:read('*a')) or nil)
   if F then F:close() end
   if not T then
      error('Could not find template file '..TemplatePath)
   end
   Env = ltp.merge_index(Env, _G)
   return ltp.render_template(T, '<?lua', '?>', Env)
end

local function indexPageCachePath(ProjectGuid)
   return ifdoc.SANDBOX_DIR..'\\'..ProjectGuid..'.report.cache.html'
end

local function createTemplateEnv()
   local Env = {}
   Env.htmlEscape = filter.html.enc
   Env.urlEscape = filter.uri.enc
   return Env
end

local function indexPage(Params)
   if not Params.project then
      local Env = createTemplateEnv()
      
      Env.ProjectList = ifdoc.iguana.projectList()
      
      local Body = renderTemplateFile('index.html', Env)
      
      net.http.respond{entity_type='text/html', body=Body}
   else
      local ProjectGuid = getRequiredParameter(Params, 'project')
      
      local Body
      
      local CacheFile =
      io.open(indexPageCachePath(ProjectGuid), 'rb')
      if CacheFile then
         Body = CacheFile:read('*a')
         CacheFile:close()
      else
         Body = ifdoc.page.generateReport(ProjectGuid)
      end
      
      net.http.respond{entity_type='text/html', body=Body}
   end
end

local AjaxHandlers = {
   ['generate']=generateData,
   ['progress']=getProgress
}

local ContentTypeMap = {
   ['.js']  = 'application/x-javascript',
   ['.css'] = 'text/css',
   ['.gif'] = 'image/gif'
}

local function contentTypeFromFileName(FileName)
   local Ext = FileName:match('.*(%.%a+)$')
   if ContentTypeMap[Ext] then
      return ContentTypeMap[Ext]
   else
      return 'text/plain'
   end
end

local function serveWebDoc(Params)
   if Params.resource then
      local ProjectFiles = iguana.project.files()
      if ProjectFiles[Params.resource] then
         local F = io.open(ProjectFiles[Params.resource], 'rb')
         local WD = ((F and F:read('*a')) or nil)
         if F then F:close() end
         local T = contentTypeFromFileName(Params.resource)
         local HttpResponse = net.http.respond{entity_type=T, body=WD}
         iguana.logDebug(HttpResponse)
         return true
      else
         net.http.respond{entity_type='text/plain', body=
            'The URL requested was not found.', code=404}
      end
   end
   return false
end

function ifdoc.page.generateReport(ProjectGuid)
   ifdoc.init(ProjectGuid)
   
   local Env = createTemplateEnv()
   Env.ProjectGuid = ProjectGuid
   Env.ProjectName = ifdoc.iguana.getProjectName(ProjectGuid) or 'Unknown project'
   
   TrackedTimes = {}
   
   local GetMessagesStart = os.time()
   Env.Messages = ifdoc.report.getMessages()
   trackTime('getMessages()', GetMessagesStart)
   
   Env.CustomProperties = ifdoc.report.getCustomProperties()
   
   local RenderStart = os.time()
   local Body = renderTemplateFile('report.html', Env)
   trackTime('renderTemplateFile()', RenderStart)
   print(TrackedTimes)
   
   local CacheFile =
   io.open(indexPageCachePath(ProjectGuid), 'w+b')
   if CacheFile then
      CacheFile:write(Body)
      CacheFile:close()
   end
   
   return Body
end

function ifdoc.page.handleRequest(Data)
   local R = net.http.parseRequest{data=Data}
   
   if serveWebDoc(R.params) then return end
   
   if AjaxHandlers[R.params.ajax] then
      local Handler = AjaxHandlers[R.params.ajax]
      if iguana.isTest() then
         Handler(R.params)
      else
         local Success, Err = pcall(Handler, R.params)
         if not Success then
            net.http.respond{entity_type='text/json',
               body=json.serialize{data={ErrorMessage=Err}}}
         end
      end
      return
   end
   
   local Handler = indexPage
   if iguana.isTest() then
      Handler(R.params)
   else
      local Success, Err = pcall(Handler, R.params)
      if not Success then
         net.http.respond{entity_type='text/html', body=
            '<html><head><title>Error</title></head><body>'..
            '<div class="err">'..filter.html.enc(Err)..
            '</div></body></html>', code=500}
      end
   end
end
