local ThisMain

local ThisProjectGuid = iguana.project.guid

local OriginalHl7Message = hl7.message
local OriginalDbTables   = db.tables
local OriginalQueuePush  = queue.push

local CurrentSampleData

-- Indexed by sample data index.
-- Values are virtual VMD paths used to generate
-- outbound messages.
local OutboundVmds = {}

-- Creates a copy of the parameters table, with corrected
-- VMD paths.
local function paramsWithVirtualVmd(Params, VirtualOtherDir)
   local VirtualParams = {}
   for K,V in pairs(Params) do
      VirtualParams[K] = V
   end
   if VirtualParams.vmd and VirtualParams.vmd ~= ifdoc.VMD then
      VirtualParams.vmd = VirtualOtherDir..'/'..VirtualParams.vmd
   end
   return VirtualParams
end

-- This converts THIS project's environment into a temporary
-- virtual environment in which the desired project can be run.
local function initVirtualProjectEnvironment(ProjectGuid)
   local ProjectDir = ifdoc.projectDir(ProjectGuid)
   local VirtualOtherDir = ProjectGuid
   
   iguana.project.guid = function() return ProjectGuid end
   ifdoc.PROJECT_GUID = ProjectGuid
   
   -- Override require() so that it loads the correct
   -- shared modules.
   require = function(Mod)
      return dofile(ProjectDir..'\\shared\\'..Mod..'.lua')
   end
   
   ThisMain = main
   main = nil -- clear global main()
   
   hl7.parse = function(Params)
      Params = paramsWithVirtualVmd(Params, VirtualOtherDir)
      local ParseResults = {ifdoc.hl7Parse(Params)}
      ifdoc.logInMessage(ParseResults[1], CurrentSampleData)
      return unpack(ParseResults)
   end
   
   hl7.message = function(Params)
      Params = paramsWithVirtualVmd(Params, VirtualOtherDir)
      OutboundVmds[CurrentSampleData] = Params.vmd
      return OriginalHl7Message(Params)
   end
   
   db.tables = function(Params)
      Params = paramsWithVirtualVmd(Params, VirtualOtherDir)
      return OriginalDbTables(Params)
   end
   
   queue.push = function(Params)
      local Vmd = OutboundVmds[CurrentSampleData]
      ifdoc.logOutMessage(Params.data, Vmd, CurrentSampleData)
      -- Hopefully the script doesn't rely on the return
      -- value of queue.push(), because we can't simulate that.
   end
end

-- Restores THIS project's environment.
local function restoreEnvironment()
   main = ThisMain -- restore original main()
   
   iguana.project.guid = ThisProjectGuid
   
   hl7.parse   = ifdoc.hl7Parse
   hl7.message = OriginalHl7Message
   db.tables   = OriginalDbTables
   queue.push  = OriginalQueuePush
end

-- Pulls the sample data from the sqlite db in which it is stored.
function ifdoc.getSampleData(ProjectGuid)
   local SampleDataDb = iguana.workingDir()..'\\data\\'..ProjectGuid..'.db'
   return db.query{api=db.SQLITE, name=SampleDataDb, live=true,
      sql='SELECT Message FROM Log'}
end

local VirtualMains = {}

-- Will initialize the "virtual project environment" for the specified
-- project.  Will return the "main" function for the project.
function ifdoc.initVirtualProjectEnv(ProjectGuid, SampleDataIndex)
   
   --if VirtualMains[ProjectGuid] then return VirtualMains[ProjectGuid] end
   
   ifdoc.checkout(ProjectGuid)
   CurrentSampleData = SampleDataIndex
   
   initVirtualProjectEnvironment(ProjectGuid)
   
   local ProjectDir = ifdoc.projectDir(ProjectGuid)
   local MainPath = ProjectDir..'\\'..ProjectGuid..'\\main.lua'
   
   local OverallSuccess, OverallError = pcall(function()
         -- Run (or "compile") the main script.
         dofile(MainPath) -- should define main()
         
         if not main then
            error(ifdoc.iguana.getProjectName(ProjectGuid)..
               ' does not define a main function.\n'..
               'Either no milestone has been saved, or main() is not defined.')
         else
            VirtualMains[ProjectGuid] = main
            main = ThisMain
         end
      end)
   
   if not OverallSuccess then
      error(OverallError)
   else
      return VirtualMains[ProjectGuid]
   end
end
